import os
from os.path import exists
import pandas as pd
from . import basic
from .amino_acids import amino_acids
from .tcr_distances_blosum import blosum
from . import translation

cdrs_sep = ';'
gap_character = '.'

all_genes = {}

class TCR_Gene:
    def __init__( self, l ):
        # l comes from pandas dataframe.itertuples
        self.id = l.id
        self.organism = l.organism
        self.chain = l.chain
        self.region = l.region
        self.nucseq = l.nucseq
        self.alseq = l.aligned_protseq
        if pd.isna(l.cdrs):
            self.cdrs = []
            self.cdr_columns = []
        else:
            self.cdrs = l.cdrs.split(cdrs_sep)
            ## these are still 1-indexed !!!!!!!!!!!!!!
            self.cdr_columns = [ list(map( int,x.split('-'))) for x in l.cdr_columns.split(cdrs_sep) ]
        frame = l.frame
        #assert frame in ['+1','+2','+3','1','2','3']
        assert frame in [1,2,3] # now parsed by pandas, so string converted to int
        #self.nucseq_offset = int( frame[-1] )-1 ## 0, 1 or 2 (0-indexed for python)
        self.nucseq_offset = frame-1 ## 0, 1 or 2 (0-indexed for python)
        self.protseq = translation.get_translation( self.nucseq, f'+{frame}' )
        assert self.protseq == self.alseq.replace(gap_character,'')
        # sanity check
        if self.cdrs:
            assert self.cdrs == [ self.alseq[ x[0]-1 : x[1] ] for x in self.cdr_columns ]

def trim_allele_to_gene( id ):
    return id[: id.index('*') ] #will fail if id doesn't contain '*'

db_file = os.path.dirname(os.path.realpath(__file__))+'/db/'+basic.db_file
assert exists(db_file)

df = pd.read_csv(db_file, sep='\t')

for l in df.itertuples():
    g = TCR_Gene( l )
    if g.organism not in all_genes:
        all_genes[g.organism] = {} # map from id to TCR_Gene objects
    all_genes[g.organism][g.id] = g


verbose = ( __name__ == '__main__' )

for organism,genes in all_genes.items():

    for ab in 'AB':
        org_merged_loopseqs = {}
        for id,g in genes.items():
            if g.chain == ab and g.region == 'V':
                loopseqs = g.cdrs[:-1] ## exclude CDR3 Nterm
                org_merged_loopseqs[id] = ' '.join( loopseqs )

        all_loopseq_nbrs = {}
        all_loopseq_nbrs_mm1 = {}
        for id1,seq1 in org_merged_loopseqs.items():
            g1 = genes[id1]
            cpos = g1.cdr_columns[-1][0] - 1 #0-indexed
            alseq1 = g1.alseq
            minlen = cpos+1
            assert len(alseq1) >= minlen
            if alseq1[cpos] != 'C' and verbose:
                print('funny cpos:',id1,alseq1,g1.cdrs[-1])

            all_loopseq_nbrs[id1] = []
            all_loopseq_nbrs_mm1[id1] = []
            for id2,seq2 in org_merged_loopseqs.items():
                g2 = genes[id2]
                alseq2 = g2.alseq
                assert len(alseq2) >= minlen
                assert len(seq1) == len(seq2)
                if seq1 == seq2:
                    all_loopseq_nbrs[id1].append( id2 )
                    all_loopseq_nbrs_mm1[id1].append( id2 )
                    continue

                ## count mismatches between these two, maybe count as an "_mm1" nbr
                loop_mismatches = 0
                loop_mismatches_cdrx = 0
                loop_mismatch_seqs =[]
                spaces=0
                for a,b in zip( seq1,seq2):
                    if a==' ':
                        spaces+=1
                        continue
                    if a!= b:
                        if a in '*.' or b in '*.':
                            loop_mismatches += 10
                            break
                        else:
                            if not (a in amino_acids and b in amino_acids):
                                print( id1,id2,a,b)
                            assert a in amino_acids and b in amino_acids
                            if spaces<=1:
                                loop_mismatches += 1
                                loop_mismatch_seqs.append( ( a,b ) )
                            else:
                                assert spaces==2
                                loop_mismatches_cdrx += 1
                            if loop_mismatches>1:
                                break
                if loop_mismatches <=1:
                    all_mismatches = 0
                    for a,b in zip( alseq1[:cpos+2],alseq2[:cpos+2]):
                        if a!= b:
                            if a in '*.' or b in '*.':
                                all_mismatches += 10
                            else:
                                if not (a in amino_acids and b in amino_acids):
                                    print( id1,id2,a,b)
                                assert a in amino_acids and b in amino_acids
                                all_mismatches += 1
                    #dist = tcr_distances.blosum_sequence_distance( seq1, seq2, gap_penalty=10 )
                    if loop_mismatches<=1 and loop_mismatches + loop_mismatches_cdrx <= 2 and all_mismatches<=10:
                        if loop_mismatches == 1:
                            blscore= blosum[(loop_mismatch_seqs[0][0],loop_mismatch_seqs[0][1])]
                        else:
                            blscore = 100
                        if blscore>=1:
                            all_loopseq_nbrs_mm1[id1].append( id2 )
                            if loop_mismatches>0 and verbose:
                                mmstring = ','.join(['%s/%s'%(x[0],x[1]) for x in loop_mismatch_seqs])
                                gene1 = trim_allele_to_gene( id1 )
                                gene2 = trim_allele_to_gene( id2 )
                                if gene1 != gene2 and verbose:
                                    print('v_mismatches:',organism,mmstring,blscore,id1,id2,\
                                        loop_mismatches,loop_mismatches_cdrx,all_mismatches,seq1)
                                    print('v_mismatches:',organism,mmstring,blscore,id1,id2,\
                                        loop_mismatches,loop_mismatches_cdrx,all_mismatches,seq2)


        for id in all_loopseq_nbrs:
            rep = min( all_loopseq_nbrs[id] )
            assert org_merged_loopseqs[id] == org_merged_loopseqs[ rep ]
            genes[id].rep = rep
            if verbose:
                print('vrep %s %15s %15s %s'%(organism, id, rep, org_merged_loopseqs[id]))


        ## merge mm1 nbrs to guarantee transitivity
        while True:
            new_nbrs = False
            for id1 in all_loopseq_nbrs_mm1:
                new_id1_nbrs = False
                for id2 in all_loopseq_nbrs_mm1[id1]:
                    for id3 in all_loopseq_nbrs_mm1[id2]:
                        if id3 not in all_loopseq_nbrs_mm1[id1]:
                            all_loopseq_nbrs_mm1[id1].append( id3 )
                            if verbose:
                                print('new_nbr:',id1,'<--->',id2,'<--->',id3)
                            new_id1_nbrs = True
                            break
                    if new_id1_nbrs:
                        break
                if new_id1_nbrs:
                    new_nbrs = True
            if verbose:
                print('new_nbrs:',ab,organism,new_nbrs)
            if not new_nbrs:
                break

        for id in all_loopseq_nbrs_mm1:
            rep = min( all_loopseq_nbrs_mm1[id] )
            genes[id].mm1_rep = rep
            if verbose:
                print('mm1vrep %s %15s %15s %s'%(organism, id, rep,org_merged_loopseqs[id]))


    ## setup Jseq reps
    for ab in 'AB':
        jloopseqs = {}
        for id,g in genes.items():
            if g.chain == ab and g.region == 'J':
                num = len( g.cdrs[0].replace( gap_character, '' ) )
                jloopseq = g.protseq[:num+3] ## go all the way up to and including the GXG
                jloopseqs[id] = jloopseq
        all_jloopseq_nbrs = {}
        for id1,seq1 in jloopseqs.items():
            all_jloopseq_nbrs[id1] = []
            for id2,seq2 in jloopseqs.items():
                if seq1 == seq2:
                    all_jloopseq_nbrs[id1].append( id2 )
        for id in all_jloopseq_nbrs:
            rep = min( all_jloopseq_nbrs[id] )
            genes[id].rep = rep
            genes[id].mm1_rep = rep # just so we have an mm1_rep field defined...
            assert jloopseqs[id] == jloopseqs[ rep ]
            if verbose:
                print('jrep %s %15s %15s %15s'%(organism, id, rep, jloopseqs[id]))



    ## setup a mapping that we can use for counting when allowing mm1s and also ignoring alleles

    # allele2mm1_rep_gene_for_counting = {}
    # def get_mm1_rep_ignoring_allele( gene, organism ): # helper fxn
    #     rep = get_mm1_rep( gene, organism )
    #     rep = rep[:rep.index('*')]
    #     return rep

    #allele2mm1_rep_gene_for_counting[ organism ] = {}

    if not basic.CLASSIC_COUNTREPS:
        # simpler scheme for choosing the 'count_rep' field
        for id, g in all_genes[organism].items():
            g.count_rep = trim_allele_to_gene(id)
    else:
        for chain in 'AB':
            for vj in 'VJ':
                allele_gs = [ (id,g) for (id,g) in all_genes[organism].items() if g.chain==chain and g.region==vj]

                gene2rep = {}
                gene2alleles = {}
                rep_gene2alleles = {}

                for allele,g in allele_gs:
                    #assert allele[2] == chain
                    gene = trim_allele_to_gene( allele )
                    rep_gene = trim_allele_to_gene( g.mm1_rep )
                    if rep_gene not in rep_gene2alleles:
                        rep_gene2alleles[ rep_gene ] = []
                    rep_gene2alleles[ rep_gene ].append( allele )

                    if gene not in gene2rep:
                        gene2rep[gene] = set()
                        gene2alleles[gene] = []
                    gene2rep[ gene ].add( rep_gene )
                    gene2alleles[gene].append( allele )

                merge_rep_genes = {}
                for gene,reps in gene2rep.items():
                    if len(reps)>1:
                        if verbose:
                            print('multireps:',organism, gene, reps)
                            for allele in gene2alleles[gene]:
                                print(' '.join(all_genes[organism][allele].cdrs), allele, \
                                    all_genes[organism][allele].rep, \
                                    all_genes[organism][allele].mm1_rep)
                        assert vj=='V'

                        ## we are going to merge these reps
                        ## which one should we choose?
                        l = [ (len(rep_gene2alleles[rep]), rep ) for rep in reps ]
                        l.sort()
                        l.reverse()
                        assert l[0][0] > l[1][0]
                        toprep = l[0][1]
                        for (count,rep) in l:
                            if rep in merge_rep_genes:
                                # ACK need to think more about this, should probably just kill this logic!
                                assert rep == toprep and merge_rep_genes[rep] == rep
                            merge_rep_genes[ rep ] = toprep


                for allele,g in allele_gs:
                    count_rep = trim_allele_to_gene( g.mm1_rep ) #get_mm1_rep_ignoring_allele( allele, organism )
                    if count_rep in merge_rep_genes:
                        count_rep = merge_rep_genes[ count_rep ]
                    g.count_rep = count_rep #allele2mm1_rep_gene_for_counting[ organism ][ allele] = count_rep
                    if verbose:
                        print('countrep:',organism, allele, count_rep)


if __name__ == '__main__':
    for org, genes in all_genes.items():
        for id, g in genes.items():
            print( org, id, g.cdrs )




from setuptools import setup, find_packages
import versioneer
PACKAGES = find_packages()

# read the contents of your README file
from os import path
this_directory = path.abspath(path.dirname(__file__))
with open(path.join(this_directory, 'README.md'), encoding='utf-8') as f:
    long_description = f.read()

opts = dict(name='snetcr',
            maintainer='Sebastiaan Valkiers',
            maintainer_email='sebastiaan.valkiers@uantwerpen.be',
            description='Pairwise distance calculation and neighborhood enrichment analysis of TCR repertoires',
            long_description=long_description,
            long_description_content_type='text/markdown',
            url='https://github.com/svalkiers/immune_response_detection',
            license='MIT',
            author='Sebastiaan Valkiers',
            author_email='sebastiaan.valkiers@uantwerpen.be',
            version=versioneer.get_version(),
            cmdclass=versioneer.get_cmdclass(),
            packages=PACKAGES,
            python_requires='>=3.9',
            classifiers=[
                'Programming Language :: Python :: 3',
                'Programming Language :: Python :: 3.9',
                'License :: OSI Approved :: MIT License',
                'Operating System :: OS Independent',
            ],
            include_package_data=True)

install_reqs = [
      'numpy>=1.23.5',
      'pandas>=1.5.2',
      'faiss-cpu==1.7.3',
      'scipy==1.9.3', 
      'scikit-learn==1.2.0',
      'pynndescent==0.5.8',
      'igraph==0.10.3',
      'networkx==3.0',
      'olga==1.2.4',
      'leidenalg==0.9.1',
      'logomaker==0.8',
      'statsmodels==0.14.0',
      'matplotlib>=3.6.2',
      'pip'
      ]

if __name__ == "__main__":
      setup(**opts, install_requires=install_reqs)


######################################################################################88

from .basic import *
from .all_genes import all_genes
#import parse_tsv
from collections import Counter
from itertools import chain
import sys
import pandas as pd
#from ..util import organism2vdj_type, IG_VDJ_TYPE
# different types of repertoire data we might have
TCR_AB_VDJ_TYPE = 'TCR_AB_VDJ_TYPE'
TCR_GD_VDJ_TYPE = 'TCR_GD_VDJ_TYPE'
IG_VDJ_TYPE = 'IG_VDJ_TYPE'

organism2vdj_type = {
    'human':TCR_AB_VDJ_TYPE,
    'mouse':TCR_AB_VDJ_TYPE,
    'human_gd':TCR_GD_VDJ_TYPE,
    'mouse_gd':TCR_GD_VDJ_TYPE,
    'human_ig':IG_VDJ_TYPE,
    'mouse_ig':IG_VDJ_TYPE,
}

MIN_CDR3_LEN = 6 # otherwise tcrdist barfs; actually we could also set this to 5 and be OK

def show(tcr):
    "For debugging"
    if type( tcr[0] ) is str:
        return ' '.join(tcr[:3])
    else:
        return ' '.join( show(x) for x in tcr )

def fixup_gene_name( gene, gene_suffix, expected_gene_names ):

    if gene in expected_gene_names:
        return gene # ALL DONE

    if '*' not in gene:
        gene += gene_suffix

    if gene in expected_gene_names:
        return gene # ALL DONE

    vj = gene[3]
    if vj not in 'VJ':
        print('wonky gene name:', gene)
        return gene

    if vj=='V' and 'DV' in gene:
        # sometimes they just delete the '/'
        # sometimes they convert it into a '-'
        new_gene = gene[:gene.index('DV')]+'/'+gene[gene.index('DV'):]
        if new_gene in expected_gene_names:
            gene = new_gene
        else:
            new_gene = gene[:gene.index('DV')-1]+'/'+gene[gene.index('DV'):]
            if new_gene in expected_gene_names:
                gene = new_gene
            else:
                print('trouble parsing V gene with DV in it:', gene)

    return gene # may still not be in expected_gene_names, will check for that later

def get_ab_from_10x_chain(chain, organism):
    ''' Returns None if the chain is not valid for this 'organism'
    '''
    if organism in ['human', 'mouse']:
        if chain in ['TRA','TRB']:
            return chain[2]
        else:
            return None
    elif organism in ['human_gd','mouse_gd']:
        if chain in ['TRA','TRG','TRD']:
            return 'A' if chain=='TRG' else 'B'
        else:
            return None
    elif organism in ['human_ig','mouse_ig']:
        if chain in ['IGH', 'IGK', 'IGL']:
            return 'B' if chain=='IGH' else 'A'
        else:
            return None
    else:
        print('unrecognized organism in get_ab_from_10x_chain:', organism)
        sys.exit()


def read_tcr_data(
        organism,
        contig_annotations_csvfile,
        consensus_annotations_csvfile,
        allow_unknown_genes = False,
        verbose = False,
        prefix_clone_ids_with_tcr_type = False,
):
    """ Parse tcr data, only taking 'productive' tcrs

    Returns:

    clonotype2tcrs, clonotype2barcodes

    """

    expected_gene_names = set(all_genes[organism].keys())

    #from cdr3s_human import all_align_fasta

    gene_suffix = '*01' # may not be used


    if prefix_clone_ids_with_tcr_type:
        if organism2vdj_type[organism] == IG_VDJ_TYPE:
            clone_id_prefix = 'bcr_'
        else:
            clone_id_prefix = 'tcr_'
    else:
        clone_id_prefix = ''

    # read the contig annotations-- map from clonotypes to barcodes
    # barcode,is_cell,contig_id,high_confidence,length,chain,v_gene,d_gene,j_gene,c_gene,full_length,productive,cdr3,cdr3_nt,reads,umis,raw_clonotype_id,raw_consensus_id
    # AAAGATGGTCTTCTCG-1,True,AAAGATGGTCTTCTCG-1_contig_1,True,695,TRB,TRBV5-1*01,TRBD2*02,TRBJ2-3*01,TRBC2*01,True,True,CASSPLAGYAADTQYF,TGCGCCAGCAGCCCCCTAGCGGGATACGCAGCAGATACGCAGTATTTT,9427,9,clonotype14,clonotype14_consensus_1
    assert exists( contig_annotations_csvfile )

    #_, lines = parse_csv_file(contig_annotations_csvfile)
    df = pd.read_csv(contig_annotations_csvfile)
    df['productive'] = df['productive'].astype(str) #sometimes it already is if there are 'Nones' in there...
    clonotype2barcodes = {}
    clonotype2tcrs_backup = {} ## in case we dont have a consensus_annotations_csvfile

    if verbose:
        print('read:', df.shape[0], 'lines from', contig_annotations_csvfile)

    for l in df.itertuples():
        bc = l.barcode
        clonotype = str(clone_id_prefix) + str(l.raw_clonotype_id)
        # annoying: pandas sometimes converts to True/False booleans
        #  and sometimes not.
        assert l.productive in ['None', 'False', 'True','TRUE','FALSE']
        if clonotype.lower() in ['none', 'nan']:
            if verbose:
                print('skip: clonotype==None/nan', l)
            continue
        if clonotype not in clonotype2barcodes:
            clonotype2barcodes[clonotype] = []
        if bc in clonotype2barcodes[clonotype]:
            pass
        else:
            clonotype2barcodes[clonotype].append( bc )

        ## experimenting here ########################################3
        if l.productive.lower() != 'true':
            if verbose:
                print('skip: productive!=True', l)
            continue
        if 'none' in [l.cdr3.lower(), l.cdr3_nt.lower(), l.v_gene.lower(),
                      l.j_gene.lower()]:
            if verbose:
                print('skip: none cdr3 or v/j genes', l)
            continue

        chain = l.chain
        ab = get_ab_from_10x_chain(chain, organism)
        if ab is None:
            if verbose:
                print('skip: ab=None', l)
            continue
        if clonotype not in clonotype2tcrs_backup:
            clonotype2tcrs_backup[ clonotype ] = {'A':Counter(), 'B':Counter() }
        # stolen from below
        vg = fixup_gene_name(l.v_gene, gene_suffix, expected_gene_names)
        jg = fixup_gene_name(l.j_gene, gene_suffix, expected_gene_names)

        if vg not in expected_gene_names:
            print('unrecognized V gene:', organism, vg)
            if not allow_unknown_genes:
                continue
        if jg not in expected_gene_names:
            print('unrecognized J gene:', organism, jg)
            if not allow_unknown_genes:
                continue
        #assert vg in all_align_fasta[organism]
        #assert jg in all_align_fasta[organism]

        tcr_chain = ( vg, jg, l.cdr3, l.cdr3_nt.lower() )

        clonotype2tcrs_backup[clonotype][ab][tcr_chain] += int(l.umis)

    for id in clonotype2tcrs_backup:
        for ab in 'AB':
            for t1,count1 in clonotype2tcrs_backup[id][ab].items():
                for t2, count2 in clonotype2tcrs_backup[id][ab].items():
                    if t2<=t1:continue
                    if t1[3] == t2[3]:
                        print('repeat??', count1, count2, t1, t2)



    if consensus_annotations_csvfile is None:
        # this is now the default-- we don't really need the code below
        #  and it won't be run unless user also provides a
        #  consensus_annotations_csvfile
        clonotype2tcrs = clonotype2tcrs_backup
    else:

        ## now read details on the individual chains for each clonotype
        # ==> tcr/human/JCC176_TX2_TCR_consensus_annotations.csv <==
        # clonotype_id,consensus_id,length,chain,v_gene,d_gene,j_gene,c_gene,full_length,productive,cdr3,cdr3_nt,reads,umis
        # clonotype100,clonotype100_consensus_1,550,TRB,TRBV24-1*01,TRBD1*01,TRBJ2-7*01,TRBC2*01,True,True,CATSDPGQGGYEQYF,TGTGCCACCAGTGACCCCGGACAGGGAGGATACGAGCAGTACTTC,8957,9

        assert exists(consensus_annotations_csvfile)
        df = pd.read_csv( consensus_annotations_csvfile )
        df['productive'] = df['productive'].astype(str) #sometimes it already is if there are 'Nones' in there...


        ## first get clonotypes with one alpha and one beta
        clonotype2tcrs = {}

        for l in df.itertuples():
            assert l.productive in [ 'None', 'False', 'True']
            if l.productive == 'True':
                id = l.clonotype_id
                if id not in clonotype2tcrs:
                    # dictionaries mapping from tcr to umi-count
                    clonotype2tcrs[id] = { 'A':Counter(), 'B':Counter() } #, 'G':[], 'D': [] }
                    assert id in clonotype2barcodes

                ch = l.chain
                ab = get_ab_from_10x_chain(ch, organism)
                if ab is None:
                    print('skipline:', consensus_annotations_csvfile, ch, l.v_gene, l.j_gene)
                    continue
                vg = fixup_gene_name(l.v_gene, gene_suffix, expected_gene_names)
                jg = fixup_gene_name(l.j_gene, gene_suffix, expected_gene_names)

                if vg not in expected_gene_names:
                    print('unrecognized V gene:', organism, vg)
                    if not allow_unknown_genes:
                        continue
                if jg not in expected_gene_names:
                    print('unrecognized J gene:', organism, jg)
                    if not allow_unknown_genes:
                        continue
                #assert vg in all_align_fasta[organism]
                #assert jg in all_align_fasta[organism]
                tcr_chain = ( vg, jg, l.cdr3, l.cdr3_nt.lower() )

                if tcr_chain not in clonotype2tcrs[id][ab]:
                    umis = int( l.umis )
                    clonotype2tcrs[id][ab][ tcr_chain ] = umis
                    old_umis = clonotype2tcrs_backup[id][ab][tcr_chain]
                    if umis != old_umis:
                        print('diff_umis:',umis, old_umis, id,ab,tcr_chain)
                else:
                    print('repeat?',id,ab,tcr_chain)
            # else:
            #     if l.productive not in [ 'None','False' ]:
            #         print('unproductive?',l.productive)


        if verbose:
            idl1 = sorted( clonotype2tcrs_backup.keys())
            idl2 = sorted( clonotype2tcrs.keys())
            print('same ids:', len(idl1), len(idl2), idl1==idl2)
            for id in clonotype2tcrs_backup:
                if id in clonotype2tcrs:
                    for ab in 'AB':
                        tl1 = sorted(clonotype2tcrs_backup[id][ab].keys())
                        tl2 = sorted(clonotype2tcrs[id][ab].keys())
                        if tl1 != tl2:
                            print('diffids:',id,ab,tl1,tl2)

    return clonotype2tcrs, clonotype2barcodes

def read_tcr_data_batch(
        organism,
        metadata_file,
        replace_batch_id = True,
        strip_batch_id_location = 'suffix',
        add_batch_id_location = 'suffix',
        batch_id_delim = '-',
        allow_unknown_genes = False,
        verbose = False,
        prefix_clone_ids_with_tcr_type = False,
):
    """ Parse tcr data from multiple batches

    metadata_file is a comma-separated (.csv) or tab-separated (.tsv) file
    that should have the columns 'file' and 'suffix'

    Returns:

    clonotype2tcrs, clonotype2barcodes

    """
    assert exists( metadata_file )

    if metadata_file.endswith('.csv'):
        sep = ','
    else:
        sep = '\t'
    md = pd.read_csv(metadata_file, sep=sep, dtype=str)

    if prefix_clone_ids_with_tcr_type:
        if organism2vdj_type[organism] == IG_VDJ_TYPE:
            clone_id_prefix = 'bcr_'
        else:
            clone_id_prefix = 'tcr_'
    else:
        clone_id_prefix = ''

    # read in contig files and update suffix to match GEX matrix
    contig_list = []
    if replace_batch_id:

        assert strip_batch_id_location in ('prefix','suffix'), "Specify either 'prefix' or 'suffix'"
        assert add_batch_id_location in ('prefix','suffix'), "Specify either 'prefix' or 'suffix'"
        assert batch_id_delim is not None, "Specify delimiter between batch_id and barcode"

        for x in md.itertuples():
            dfx = pd.read_csv( x.file )
            batch_id = str(x.batch_id)

            if strip_batch_id_location == 'suffix':
                barcodes = dfx['barcode'].str.split(batch_id_delim).str.get(0)
            else:
                barcodes = dfx['barcode'].str.split(batch_id_delim).str.get(1)

            if add_batch_id_location == 'suffix':
                dfx['barcode'] = barcodes + batch_id_delim + batch_id
                dfx['contig_id'] = barcodes + batch_id_delim + batch_id + '_contig_' + dfx['contig_id'].str.split('_').str.get(2) # currently unused, but can't hurt
            else:
                dfx['barcode'] = batch_id + batch_id_delim + barcodes
                dfx['contig_id'] = batch_id + batch_id_delim + barcodes + '_contig_' + dfx['contig_id'].str.split('_').str.get(2)

            dfx['raw_clonotype_id'] = clone_id_prefix + dfx['raw_clonotype_id'] + '_' + batch_id
            dfx['raw_consensus_id'] = clone_id_prefix + dfx['raw_consensus_id'] + '_' + batch_id
            contig_list.append(dfx)
    else:
        print('replace_batch_id set to False. Barcode batch_id will not be modified')
        for x in md.itertuples():
            dfx = pd.read_csv( x.file )
            contig_list.append(dfx)

    # merge filtered_contig_annotations tables
    df = pd.concat(contig_list)

    expected_gene_names = set(all_genes[organism].keys())

    gene_suffix = '*01' # may not be used

    # read the contig annotations-- map from clonotypes to barcodes
    # and from clonotypes to tcrs
    #
    # example header and first line:
    # barcode,is_cell,contig_id,high_confidence,length,chain,v_gene,d_gene,j_gene,c_gene,full_length,productive,cdr3,cdr3_nt,reads,umis,raw_clonotype_id,raw_consensus_id
    # AAAGATGGTCTTCTCG-1,True,AAAGATGGTCTTCTCG-1_contig_1,True,695,TRB,TRBV5-1*01,TRBD2*02,TRBJ2-3*01,TRBC2*01,True,True,CASSPLAGYAADTQYF,TGCGCCAGCAGCCCCCTAGCGGGATACGCAGCAGATACGCAGTATTTT,9427,9,clonotype14,clonotype14_consensus_1

    df['productive'] = df['productive'].astype(str)
    clonotype2tcrs = {}
    clonotype2barcodes = {}
    for l in df.itertuples():
        # the fields we use:   barcode  raw_clonotype_id  productive  cdr3  cdr3_nt  chain  v_gene  j_gene  umis
        bc = l.barcode
        clonotype = l.raw_clonotype_id

        assert l.productive in [ 'None', 'False', 'True']
        if clonotype =='None': # I think this includes contigs w/ is_cell=False, e.g.
            continue
        if clonotype not in clonotype2barcodes:
            clonotype2barcodes[clonotype] = []
        if bc in clonotype2barcodes[clonotype]:
            pass
        else:
            clonotype2barcodes[clonotype].append( bc )

        if l.productive != 'True':
            continue
        if l.cdr3.lower() == 'none' or l.cdr3_nt.lower() == 'none':
            continue

        chain = l.chain
        ab = get_ab_from_10x_chain(chain, organism)
        if ab is None:
            continue
        if clonotype not in clonotype2tcrs:
            clonotype2tcrs[ clonotype ] = {'A':Counter(), 'B':Counter() }

        vg = fixup_gene_name(l.v_gene, gene_suffix, expected_gene_names)
        jg = fixup_gene_name(l.j_gene, gene_suffix, expected_gene_names)

        if vg not in expected_gene_names:
            print('unrecognized V gene:', organism, vg)
            if not allow_unknown_genes:
                continue
        if jg not in expected_gene_names:
            print('unrecognized J gene:', organism, jg)
            if not allow_unknown_genes:
                continue

        tcr_chain = ( vg, jg, l.cdr3, l.cdr3_nt.lower() )

        clonotype2tcrs[clonotype][ab][tcr_chain] += int(l.umis)

    for id in clonotype2tcrs:
        for ab in 'AB':
            for t1,count1 in clonotype2tcrs[id][ab].items():
                for t2, count2 in clonotype2tcrs[id][ab].items():
                    if t2<=t1:continue
                    if t1[3] == t2[3]:
                        print('repeat??', count1, count2, t1, t2)


    return clonotype2tcrs, clonotype2barcodes

def _make_clones_file( organism, outfile, clonotype2tcrs, clonotype2barcodes, verbose=False ):
    ''' Make a clones file with information parsed from the 10X csv files

    organism is one of ['mouse','human']

    outfile is the name of the clones file to be created

    '''

    #tmpfile = outfile+'.tmp' # a temporary intermediate file
    bc_mapfile = outfile+'.barcode_mapping.tsv'
    outmap = open(bc_mapfile,'w')
    outmap.write('clone_id\tbarcodes\n')

    outfields = 'clone_id subject clone_size va_gene ja_gene va2_gene ja2_gene vb_gene jb_gene cdr3a cdr3a_nucseq cdr3a2 cdr3a2_nucseq cdr3b cdr3b_nucseq'\
        .split()
    extra_fields = 'alpha_umi alpha2_umi beta_umi num_alphas num_betas'.split()
    outfields += extra_fields

    # we used to make a temporary file and then run tcr-dist/file_converter.py on it
    # for this slim version, just take the temporary file
    #out = open(tmpfile,'w')
    out = open(outfile,'w')
    out.write('\t'.join( outfields )+'\n' )

    for clonotype in sorted(clonotype2tcrs.keys()):
        tcrs = clonotype2tcrs[clonotype]
        if len(tcrs['A']) >= 1 and len(tcrs['B']) >= 1:
            atcrs = tcrs['A'].most_common()
            btcrs = tcrs['B'].most_common()
            if len(atcrs)>1:
                if verbose:
                    print('multiple alphas, picking top umi:',' '.join( str(x) for _,x in atcrs ))
                atcr2, atcr2_umi = atcrs[1]
            else:
                atcr2, atcr2_umi = ('', '', '', ''), 0
            if len(btcrs)>1 and verbose:
                print('multiple  betas, picking top umi:',' '.join( str(x) for _,x in btcrs ))
            atcr, atcr_umi = atcrs[0]
            btcr, btcr_umi = btcrs[0]
            outl = {}
            outl['clone_id'] = clonotype
            outl['subject'] = 'UNK_S'
            outl['clone_size'] = len(clonotype2barcodes[clonotype])
            outl['va_gene']      = atcr[0]
            outl['ja_gene']      = atcr[1]
            outl['cdr3a']        = atcr[2]
            outl['cdr3a_nucseq'] = atcr[3]
            outl['alpha_umi']    = str(atcr_umi)
            outl['va2_gene']      = atcr2[0]
            outl['ja2_gene']      = atcr2[1]
            outl['cdr3a2']        = atcr2[2]
            outl['cdr3a2_nucseq'] = atcr2[3]
            outl['alpha2_umi']    = str(atcr2_umi)
            outl['num_alphas']   = str(len(atcrs))
            outl['vb_gene']      = btcr[0]
            outl['jb_gene']      = btcr[1]
            outl['cdr3b']        = btcr[2]
            outl['cdr3b_nucseq'] = btcr[3]
            outl['beta_umi']     = str(btcr_umi)
            outl['num_betas']    = str(len(btcrs))
            if len(outl['cdr3a']) < MIN_CDR3_LEN or  len(outl['cdr3b']) < MIN_CDR3_LEN:
                if verbose:
                    print('Warning: skipping clonotype with short cdr3s: {} {} {}'\
                          .format(clonotype, outl['cdr3a'], outl['cdr3b'] ))
                continue
            out.write( '\t'.join(str(outl[x]) for x in outfields)+'\n')
            outmap.write('{}\t{}\n'.format(clonotype,','.join(clonotype2barcodes[clonotype])))
    out.close()
    outmap.close()

    # for the time being, go with the clones file we just made even though it doesn't have all the usual stupid info


def setup_filtered_clonotype_dicts(
        clonotype2tcrs,
        clonotype2barcodes,
        min_repeat_count_fraction = 0.33,
        verbose = False
):
    ''' returns new_clonotype2tcrs, new_clonotype2barcodes
    '''

    # get count of how many cells support each pairing
    #  (pairing_counts_by_umi was just for diagnostics)
    pairing_counts, pairing_counts_by_umi = Counter(), Counter()
    for cid, tcrs in clonotype2tcrs.items():
        size = len(clonotype2barcodes[cid])
        for t1, umi1 in tcrs['A'].items():
            for t2, umi2 in tcrs['B'].items():
                pairing_counts[ (t1,t2) ] += size
                pairing_counts_by_umi[ (t1,t2) ] += min(umi1, umi2)

    # this is no longer going to be a bijection!
    chain_partner = {}

    valid_ab_pairings = set()

    for ( (t1,t2), count ) in list(pairing_counts.most_common()): # make a copy for sanity
        if t1 in chain_partner or t2 in chain_partner:
            t1_p2 = chain_partner.get(t1,None)
            t2_p2 = chain_partner.get(t2,None)

            oldcount1 = pairing_counts[ (t1,t1_p2) ]
            oldcount2 = pairing_counts[ (t2_p2,t2) ]

            if count >= min_repeat_count_fraction * max(oldcount1,oldcount2):
                # take it anyway -- second alpha or genuinely shared alpha?

                if verbose:
                    print('take_rep_partners:', count, oldcount1, oldcount2, t1, t2, t1_p2, t2_p2)
                # dont overwrite the old pairing... might not do either of these!
                if t1 not in chain_partner:
                    chain_partner[t1] = t2
                if t2 not in chain_partner:
                    chain_partner[t2] = t1

                valid_ab_pairings.add( (t1, t2 ) )
            else:
                if verbose:
                    print('skip_rep_partners:', count, oldcount1, oldcount2, t1, t2, t1_p2, t2_p2)

        else: # neither chain already in chain_partner
            #
            # NOTE: removed the code checking for TIES!!!!!!!!!!!
            if verbose:
                print('norep_partners:', count, t1, t2)
            chain_partner[t1] = t2
            chain_partner[t2] = t1

            valid_ab_pairings.add( ( t1, t2 ) )


    # now let's revisit the clonotypes
    pairs_tuple2clonotypes = {}
    ab_counts = Counter() # for diagnostics
    for (clone_size, cid) in reversed( sorted( (len(y), x) for x,y in clonotype2barcodes.items() ) ):
        if cid not in clonotype2tcrs:
            #print('WHOAH missing tcrs for clonotype', clone_size, cid, clonotype2barcodes[cid])
            continue
        tcrs = clonotype2tcrs[cid]
        was_good_clone = len(tcrs['A']) >= 1 and len(tcrs['B']) >= 1
        pairs = []
        for atcr in tcrs['A']:
            for btcr in tcrs['B']:
                if ( atcr,btcr ) in valid_ab_pairings:
                    pairs.append( (atcr,btcr) )
        if pairs:
            pairs_tuple = None
            if len(pairs)>1:
                alphas = set( x[0] for x in pairs )
                betas = set( x[1] for x in pairs )
                ab_counts[ (len(alphas),len(betas) ) ] += clone_size

                if len(alphas) == 2 and len(betas) == 1: ## only allow double-alphas
                    assert len(pairs) == 2
                    pairs_tuple = tuple(x[1] for x in reversed(sorted( [ (pairing_counts[x],x) for x in pairs ] )))
                    assert len( pairs_tuple)==2
                    # confirm ordered by counts
                    assert pairing_counts[pairs_tuple[0]] >= pairing_counts[pairs_tuple[1]]
            else:
                ab_counts[ (1,1) ] += clone_size
                pairs_tuple = tuple(pairs)
                assert len(pairs_tuple) == 1

            if pairs_tuple:
                pairs_tuple2clonotypes.setdefault( pairs_tuple, [] ).append( cid )
            else:
                if verbose:
                    print('SKIPCLONE:', was_good_clone, cid, clone_size, pairs, 'bad_pairs')
        else:
            if verbose:
                print('SKIPCLONE:', was_good_clone, cid, clone_size, 'no_valid_pairs')


    ## reorder pairs_tuples in the case of ties, using umis
    reorder = []
    for pt, clonotypes in pairs_tuple2clonotypes.items():
        assert len(pt) in [1,2]
        if len(pt) == 2:
            assert pt[0][1] == pt[1][1] # same beta chain
            at1, bt = pt[0]
            at2, _  = pt[1]
            count1, count2 = pairing_counts[(at1,bt)], pairing_counts[(at2,bt)]
            assert count1 >= count2
            if count1 == count2:
                # no way to figure out which is the primary alpha!
                # look at umis?
                c1 = sum( clonotype2tcrs[x]['A'][at1] for x in clonotypes )
                c2 = sum( clonotype2tcrs[x]['A'][at2] for x in clonotypes )
                if verbose:
                    print('alphatie:', count1, count2, c1, c2, show(at1), show(at2), show(bt))
                if c2 > c1:
                    reorder.append( pt )

    for pt in reorder:
        #print('reorder:', show(pt))
        assert len(pt) == 2
        rpt = (pt[1], pt[0])
        assert pt in pairs_tuple2clonotypes and rpt not in pairs_tuple2clonotypes
        pairs_tuple2clonotypes[rpt] = pairs_tuple2clonotypes[pt][:]
        del pairs_tuple2clonotypes[pt]


    ## look for len1 pairs_tuples that overlap with two different len2 pairs_tuples
    merge_into_pairs = []
    for ii, pt1 in enumerate(pairs_tuple2clonotypes):
        if len(pt1) == 1:
            if verbose and ii%500==0:
                print('look for overlaps:', ii, len(pairs_tuple2clonotypes))
            overlaps = []
            for pt2 in pairs_tuple2clonotypes:
                if len(pt2) == 2 and pt2[0] == pt1[0]:
                    overlaps.append( pt2 )
                elif len(pt2) == 2 and pt2[1] == pt1[0]:
                    # this 'elif' was added 2020-12-12; before that we were missing some
                    # overlaps...
                    # print('MISSED OVERLAP:', pairing_counts[pt2[0]], pairing_counts[pt2[1]],
                    #       pairing_counts_by_umi[pt2[0]], pairing_counts_by_umi[pt2[1]],
                    #       show(pt1), show(pt2))
                    overlaps.append( pt2 )
            if len(overlaps)>1:
                if verbose:
                    print('badoverlaps:', len(overlaps), show(pt1), show(overlaps))
            elif len(overlaps)==1:
                pt2 = overlaps[0]
                merge_into_pairs.append( (pt1,pt2) )

    for pt1, pt2 in merge_into_pairs:
        assert len(pt1) == 1 and len(pt2) == 2
        #print('mergeinto:', show(pt1), show(pt2))
        pairs_tuple2clonotypes[pt2].extend( pairs_tuple2clonotypes[pt1] )
        del pairs_tuple2clonotypes[pt1]


    ## look for pairs_tuples that will give the same clones file line
    if verbose:
        for pt1, clonotypes in pairs_tuple2clonotypes.items():
            for pt2 in pairs_tuple2clonotypes:
                if pt1 < pt2 and pt1[0] == pt2[0]:
                    print('overlap:', len(pt1), len(pt2), pt1, pt2)


    ## now setup new clonotype2tcrs, clonotype2barcodes mappings
    new_clonotype2tcrs = {}
    new_clonotype2barcodes = {}

    for pairs_tuple, clonotypes in pairs_tuple2clonotypes.items():
        c0 = clonotypes[0]
        if len(clonotypes)>1:
            if verbose:
                print('merging:', ' '.join(clonotypes))
        tcrs = {'A':Counter(), 'B':Counter()}
        for (atcr,btcr) in pairs_tuple:
            tcrs['A'][atcr] += pairing_counts[(atcr, btcr)]
            tcrs['B'][btcr] += pairing_counts[(atcr, btcr)]
        if len(pairs_tuple)==2:
            a1, a2 = pairs_tuple[0][0], pairs_tuple[1][0]
            if tcrs['A'][a1] == tcrs['A'][a2]:
                tcrs['A'][a1] += 1 # ensure the desired order
            else:
                assert tcrs['A'][a1] > tcrs['A'][a2]
        assert len(tcrs['A']) in [1,2]
        assert len(tcrs['B']) == 1
        assert c0 not in new_clonotype2tcrs
        new_clonotype2tcrs[c0] = tcrs
        new_clonotype2barcodes[c0] = list(chain( *(clonotype2barcodes[x] for x in clonotypes)))
        # print 'new:', new_clonotype2barcodes[c0]
        # for c in clonotypes:
        #     print 'old:', clonotype2barcodes[c]
        # print len(new_clonotype2barcodes[c0]), sum( len(clonotype2barcodes[x]) for x in clonotypes)
        assert len(new_clonotype2barcodes[c0]) == sum( len(clonotype2barcodes[x]) for x in clonotypes)


    print('ab_counts:', ab_counts.most_common())
    old_good_clonotypes = [ x for x,y in clonotype2tcrs.items()
                            if len(y['A']) >= 1 and len(y['B']) >= 1 ]
    old_unpaired_clonotypes = [ x for x,y in clonotype2tcrs.items()
                                if len(y['A']) < 1 or len(y['B']) < 1 ]
    old_num_unpaired_barcodes = sum( len(clonotype2barcodes[x])
                                     for x in old_unpaired_clonotypes )
    old_num_paired_barcodes = sum( len(clonotype2barcodes[x])
                                   for x in old_good_clonotypes )
    new_num_barcodes = sum(len(x) for x in new_clonotype2barcodes.values())

    paired_barcode_fraction = (new_num_barcodes/old_num_paired_barcodes
                               if old_num_paired_barcodes else 1.)


    print('old_unpaired_barcodes:', old_num_unpaired_barcodes,
          'old_paired_barcodes:', old_num_paired_barcodes,
          'new_stringent_paired_barcodes:', new_num_barcodes,
          'paired_barcode_fraction:', paired_barcode_fraction)

    if paired_barcode_fraction<0.75:
        print('WARNING:: make_10x_clones_file.setup_filtered_clonotype_dicts:',
              'paired_barcode_fraction=', paired_barcode_fraction,
              'which seems a little low (warning threshold=0.75)')

    return new_clonotype2tcrs, new_clonotype2barcodes


def make_10x_clones_file(
        filtered_contig_annotations_csvfile,
        organism,
        clones_file, # the OUTPUT file, the one we're making
        stringent = True, # dont believe the 10x clonotypes; reduce 'duplicated' and 'fake' clones
        consensus_annotations_csvfile = None,
        verbose = False,
        **kwargs
):

    clonotype2tcrs, clonotype2barcodes = read_tcr_data(
        organism,
        filtered_contig_annotations_csvfile,
        consensus_annotations_csvfile,
        verbose=verbose,
        **kwargs
    )

    if stringent:
        clonotype2tcrs, clonotype2barcodes = setup_filtered_clonotype_dicts(
            clonotype2tcrs, clonotype2barcodes )


    _make_clones_file( organism, clones_file, clonotype2tcrs,
                       clonotype2barcodes )

def make_10x_clones_file_batch(
        metadata_file,
        organism,
        clones_file,
        replace_batch_id = True,
        strip_batch_id_location = 'suffix',
        add_batch_id_location = 'suffix',
        batch_id_delim = '-',
        stringent = True,
        **kwargs
):
    """
    Aggregates multiple filtered_conting_anntotations.csv files into a single clones.tsv file for merging with
    an aggregate gex matrix in CoNGA preprocessing.

    This was primarily motivated for cases when an aggregate gex matrix is generated through cellranger aggr
    and each samples' vdj library was processed seperately through cellranger vdj, so the default behavior is to
    strip the '-1' suffix set by cellranger vdj and replace it with those specified in the metadata_file batch_id column.

    The locations of the barcode batch_id to be stripped and where the new one should be inserted relative to the barcode can be specified,
    which might be useful in cases where the gex data was merged later through other means further downstream.

    metadata_file: path to a csv file containing two columns:
        'file': the paths to the filtered_contig_annotations.csv files for each sample being merged
        'batch_id': replacement barcode suffix or prefix to match with appropriate cell barcodes in gex file.
    organism
    clones_file: output tsvfile of aggregated sample clonotypes
    replace_batch_id: boolean specifiying if batch_id in filtered_contigs_annotations.csv files will be be replaced. (Default is True)
    strip_batch_id_location: Either 'prefix' or 'suffix' specifying the location of the batch_id relative to the umi that will be stripped. (Default = 'suffix')
    add_batch_id_location: Either 'prefix' or 'suffix' specifying the location where to add the new batch_id relative to the umi. (Default = 'suffix')
    batch_id_delim: Delimiter between the barcode and the batch_id. Also used as delimiter when applying the new batch_id. (Default = '-')
    stringent: Strongly recommended. Corrects for spurious chain pairings in case where mutliple chains of the same type are detected in the same cell. (Default True)
    kwargs: passed to read_tcr_data_batch

    """

    clonotype2tcrs, clonotype2barcodes = read_tcr_data_batch( organism,
        metadata_file,
        replace_batch_id,
        strip_batch_id_location,
        add_batch_id_location,
        batch_id_delim,
        **kwargs )

    if stringent:
        clonotype2tcrs, clonotype2barcodes = setup_filtered_clonotype_dicts( clonotype2tcrs, clonotype2barcodes )


    _make_clones_file( organism, clones_file, clonotype2tcrs, clonotype2barcodes )





# long story
CB_RED    = 'red'
CB_GREEN  = 'green'
CB_BLUE   = 'blue'
CB_ORANGE = 'orange'
CB_PURPLE = 'purple'
CB_BROWN  = 'brown'


colors_string = """AliceBlue 	#F0F8FF	 	Shades	Mix
AntiqueWhite 	#FAEBD7	 	Shades	Mix
Aqua 	#00FFFF	 	Shades	Mix
Aquamarine 	#7FFFD4	 	Shades	Mix
Azure 	#F0FFFF	 	Shades	Mix
Beige 	#F5F5DC	 	Shades	Mix
Bisque 	#FFE4C4	 	Shades	Mix
Black 	#000000	 	Shades	Mix
BlanchedAlmond 	#FFEBCD	 	Shades	Mix
Blue 	#0000FF	 	Shades	Mix
BlueViolet 	#8A2BE2	 	Shades	Mix
Brown 	#A52A2A	 	Shades	Mix
BurlyWood 	#DEB887	 	Shades	Mix
CadetBlue 	#5F9EA0	 	Shades	Mix
Chartreuse 	#7FFF00	 	Shades	Mix
Chocolate 	#D2691E	 	Shades	Mix
Coral 	#FF7F50	 	Shades	Mix
CornflowerBlue 	#6495ED	 	Shades	Mix
Cornsilk 	#FFF8DC	 	Shades	Mix
Crimson 	#DC143C	 	Shades	Mix
Cyan 	#00FFFF	 	Shades	Mix
DarkBlue 	#00008B	 	Shades	Mix
DarkCyan 	#008B8B	 	Shades	Mix
DarkGoldenRod 	#B8860B	 	Shades	Mix
DarkGray 	#A9A9A9	 	Shades	Mix
DarkGreen 	#006400	 	Shades	Mix
DarkKhaki 	#BDB76B	 	Shades	Mix
DarkMagenta 	#8B008B	 	Shades	Mix
DarkOliveGreen 	#556B2F	 	Shades	Mix
DarkOrange 	#FF8C00	 	Shades	Mix
DarkOrchid 	#9932CC	 	Shades	Mix
DarkRed 	#8B0000	 	Shades	Mix
DarkSalmon 	#E9967A	 	Shades	Mix
DarkSeaGreen 	#8FBC8F	 	Shades	Mix
DarkSlateBlue 	#483D8B	 	Shades	Mix
DarkSlateGray 	#2F4F4F	 	Shades	Mix
DarkTurquoise 	#00CED1	 	Shades	Mix
DarkViolet 	#9400D3	 	Shades	Mix
DeepPink 	#FF1493	 	Shades	Mix
DeepSkyBlue 	#00BFFF	 	Shades	Mix
DimGray 	#696969	 	Shades	Mix
DodgerBlue 	#1E90FF	 	Shades	Mix
FireBrick 	#B22222	 	Shades	Mix
FloralWhite 	#FFFAF0	 	Shades	Mix
ForestGreen 	#228B22	 	Shades	Mix
Fuchsia 	#FF00FF	 	Shades	Mix
Gainsboro 	#DCDCDC	 	Shades	Mix
GhostWhite 	#F8F8FF	 	Shades	Mix
Gold 	#FFD700	 	Shades	Mix
GoldenRod 	#DAA520	 	Shades	Mix
Gray 	#808080	 	Shades	Mix
Green 	#008000	 	Shades	Mix
GreenYellow 	#ADFF2F	 	Shades	Mix
HoneyDew 	#F0FFF0	 	Shades	Mix
HotPink 	#FF69B4	 	Shades	Mix
IndianRed  	#CD5C5C	 	Shades	Mix
Indigo  	#4B0082	 	Shades	Mix
Ivory 	#FFFFF0	 	Shades	Mix
Khaki 	#F0E68C	 	Shades	Mix
Lavender 	#E6E6FA	 	Shades	Mix
LavenderBlush 	#FFF0F5	 	Shades	Mix
LawnGreen 	#7CFC00	 	Shades	Mix
LemonChiffon 	#FFFACD	 	Shades	Mix
LightBlue 	#ADD8E6	 	Shades	Mix
LightCoral 	#F08080	 	Shades	Mix
LightCyan 	#E0FFFF	 	Shades	Mix
LightGoldenRodYellow 	#FAFAD2	 	Shades	Mix
LightGray 	#D3D3D3	 	Shades	Mix
LightGreen 	#90EE90	 	Shades	Mix
LightPink 	#FFB6C1	 	Shades	Mix
LightSalmon 	#FFA07A	 	Shades	Mix
LightSeaGreen 	#20B2AA	 	Shades	Mix
LightSkyBlue 	#87CEFA	 	Shades	Mix
LightSlateGray 	#778899	 	Shades	Mix
LightSteelBlue 	#B0C4DE	 	Shades	Mix
LightYellow 	#FFFFE0	 	Shades	Mix
Lime 	#00FF00	 	Shades	Mix
LimeGreen 	#32CD32	 	Shades	Mix
Linen 	#FAF0E6	 	Shades	Mix
Magenta 	#FF00FF	 	Shades	Mix
Maroon 	#800000	 	Shades	Mix
MediumAquaMarine 	#66CDAA	 	Shades	Mix
MediumBlue 	#0000CD	 	Shades	Mix
MediumOrchid 	#BA55D3	 	Shades	Mix
MediumPurple 	#9370DB	 	Shades	Mix
MediumSeaGreen 	#3CB371	 	Shades	Mix
MediumSlateBlue 	#7B68EE	 	Shades	Mix
MediumSpringGreen 	#00FA9A	 	Shades	Mix
MediumTurquoise 	#48D1CC	 	Shades	Mix
MediumVioletRed 	#C71585	 	Shades	Mix
MidnightBlue 	#191970	 	Shades	Mix
MintCream 	#F5FFFA	 	Shades	Mix
MistyRose 	#FFE4E1	 	Shades	Mix
Moccasin 	#FFE4B5	 	Shades	Mix
NavajoWhite 	#FFDEAD	 	Shades	Mix
Navy 	#000080	 	Shades	Mix
OldLace 	#FDF5E6	 	Shades	Mix
Olive 	#808000	 	Shades	Mix
OliveDrab 	#6B8E23	 	Shades	Mix
Orange 	#FFA500	 	Shades	Mix
OrangeRed 	#FF4500	 	Shades	Mix
Orchid 	#DA70D6	 	Shades	Mix
PaleGoldenRod 	#EEE8AA	 	Shades	Mix
PaleGreen 	#98FB98	 	Shades	Mix
PaleTurquoise 	#AFEEEE	 	Shades	Mix
PaleVioletRed 	#DB7093	 	Shades	Mix
PapayaWhip 	#FFEFD5	 	Shades	Mix
PeachPuff 	#FFDAB9	 	Shades	Mix
Peru 	#CD853F	 	Shades	Mix
Pink 	#FFC0CB	 	Shades	Mix
Plum 	#DDA0DD	 	Shades	Mix
PowderBlue 	#B0E0E6	 	Shades	Mix
Purple 	#800080	 	Shades	Mix
RebeccaPurple 	#663399	 	Shades	Mix
Red 	#FF0000	 	Shades	Mix
RosyBrown 	#BC8F8F	 	Shades	Mix
RoyalBlue 	#4169E1	 	Shades	Mix
SaddleBrown 	#8B4513	 	Shades	Mix
Salmon 	#FA8072	 	Shades	Mix
SandyBrown 	#F4A460	 	Shades	Mix
SeaGreen 	#2E8B57	 	Shades	Mix
SeaShell 	#FFF5EE	 	Shades	Mix
Sienna 	#A0522D	 	Shades	Mix
Silver 	#C0C0C0	 	Shades	Mix
SkyBlue 	#87CEEB	 	Shades	Mix
SlateBlue 	#6A5ACD	 	Shades	Mix
SlateGray 	#708090	 	Shades	Mix
Snow 	#FFFAFA	 	Shades	Mix
SpringGreen 	#00FF7F	 	Shades	Mix
SteelBlue 	#4682B4	 	Shades	Mix
Tan 	#D2B48C	 	Shades	Mix
Teal 	#008080	 	Shades	Mix
Thistle 	#D8BFD8	 	Shades	Mix
Tomato 	#FF6347	 	Shades	Mix
Turquoise 	#40E0D0	 	Shades	Mix
Violet 	#EE82EE	 	Shades	Mix
Wheat 	#F5DEB3	 	Shades	Mix
White 	#FFFFFF	 	Shades	Mix
WhiteSmoke 	#F5F5F5	 	Shades	Mix
Yellow 	#FFFF00	 	Shades	Mix
YellowGreen 	#9ACD32	 	Shades	Mix"""

html_colors = []
html_colors_no_whites = []
html_colors_no_lights = []
all_rgb = {}

def rgb_from_hex( hex ):
    assert hex[0] == '#' and len(hex)==7
    return ( int(hex[1:3],16),
             int(hex[3:5],16),
             int(hex[5:7],16) )

for line in colors_string.split('\n'):
    l = line.split()
    colorname = l[0].lower()
    if colorname == 'rebeccapurple': continue ## matplotlib doesnt like this one...
    hex = l[1]

    r,g,b = rgb_from_hex( hex )

    lowername = colorname.lower()
    html_colors.append( lowername )
    all_rgb[ lowername ] = (r,g,b)

    avg_rgb= sum((r,g,b))/3.0


    if 'white' not in lowername:
        html_colors_no_whites.append( lowername )

    if 'white' not in lowername and avg_rgb<210:
        html_colors_no_lights.append( lowername )


## get some colors for objects sorted by frequency
## fiddle with random seed/state to get consistent colors; legacy silliness...
##
import random
initial_random_state = random.getstate()

random.seed(2)

starter_colors = ['red',
                  'green',
                  'blue',
                  'cyan',
                  'magenta',
                  'black',
                  'lime',
                  'purple',
                  'gold',
                  'olive',
                  'maroon',
]
for c in starter_colors:
    assert c in html_colors_no_whites

other_colors = [ x for x in html_colors_no_whites if x not in starter_colors ]
other_colors_no_lights = [ x for x in html_colors_no_lights if x not in starter_colors ]

random.shuffle( other_colors )
rank_colors = starter_colors + other_colors

random.shuffle( other_colors_no_lights )
rank_colors_no_lights = starter_colors + other_colors_no_lights

random.setstate(initial_random_state)

def get_rank_colors( num ):
    n_rank_colors = len(rank_colors)
    colors = []

    for i in range( num ):
        colors.append( rank_colors[ i%n_rank_colors ] )
    return colors

def get_rank_colors_no_lights( num ):
    n_rank_colors = len(rank_colors_no_lights)
    colors = []

    for i in range( num ):
        colors.append( rank_colors_no_lights[ i%n_rank_colors ] )
    return colors


if __name__=='__main__':
    colors = get_rank_colors(20)
    for i,col in enumerate(colors):
        print(i,col,sum(all_rgb[col]))

    colors = get_rank_colors_no_lights(20)
    for i,col in enumerate(colors):
        print(i,col,sum(all_rgb[col]))





import pandas as pd
from abc import ABC
from functools import cached_property

from .tools import profile_matrix, motif_from_profile

class TcrCollection(ABC):
    def __init__(self, df) -> None:
        # assert all required columns are present
        required_cols = self.required_cols if hasattr(self,"required_cols") else ["v_call", "junction_aa"]
        if not set(required_cols).issubset(df.columns):
            raise ValueError(
                f"Input dataframe requires following columns: {required_cols}"
            )

        self.data = df

    def __repr__(self) -> str:
        return f"TCR collection of size {len(self.data)}"

    def __iter__(self):
        for s in self.cdr3s:
            yield s

    def __getitem__(self, n):
        return self.cdr3s[n]

    def __len__(self):
        return len(self.data)

    def to_df(self) -> pd.DataFrame:
        """Convert to pandas DataFrame"""
        return self.data

    def sample(self, n:int, weight_col:str=None):
        """
        Randomly sample n TCRs.

        Parameters
        ----------
        n : int
            Number of TCRs to sample.
        weight_col : str or array-like, optional
            If not passed, each TCR has equal probability to be picked. If an
            array is passed, array values determine the probabilty of TCR at
            that index to be picked. If a string is passed, the corresponding
            column from self.data will be used as weight.
        """
        return self.__class__(df=self.data.sample(n, weights=weight_col))

    @cached_property
    def cdr3s(self):
        return self.data["junction_aa"].to_list()


class Repertoire(TcrCollection):
    """Class for storing TCR-seq information and calculating useful properties."""

    def __repr__(self) -> str:
        return f"TCR repertoire of size {len(self.data)}"


class Cluster(TcrCollection):
    """
    Class for storing Cluster-level clonotype information, and calculating its
    useful properties.
    """

    def __repr__(self) -> str:
        return f"TCR cluster of size {len(self.data)}"


    def motif(self, method="standard", cutoff=0.7) -> str:
        return motif_from_profile(profile_matrix(self.cdr3s), method, cutoff)


class ClusteredRepertoire(TcrCollection):
    """
    Class for storing clustered clonotype information.
    """
    def __init__(self, df) -> None:
        self.cluster_ids = df.index.get_level_values(level="cluster").unique().to_list()
        super().__init__(df)

    def __repr__(self) -> str:
        if not "-1" in self.cluster_ids:
            return f"Clustered TCR repertoire of size {len(self.data)} containing {len(self.cluster_ids)} clusters."
        else:
            num_unclustered_sequences = len(self.data.xs("-1", level="cluster"))
            return f"Clustered TCR repertoire of size {len(self.data)} containing {len(self.cluster_ids)-1} clusters and {num_unclustered_sequences} unclustered sequences."

    @classmethod
    def from_clustcr_result(cls, df, clusteringresult):
        df_merged = pd.merge(left=df, right=clusteringresult.clusters_df, how="left")
        df_merged["cluster"] = df_merged["cluster"].fillna(-1).astype(int).astype(str)
        df_merged=df_merged.reset_index().set_index(["cluster", "index"])
        return cls(df_merged)

    def __iter__(self):
        for i in self.cluster_ids:
            if i == "-1":
                yield Repertoire(self.data.xs(i, level=0))
            else:
                yield Cluster(self.data.xs(i, level=0))

    def iter_clusters(self):
        for i in self.cluster_ids:
            if i != "-1":
                yield Cluster(self.data.xs(i, level=0))

    def __len__(self):
        return len(self.data)


from os import system
from os.path import exists
import base64 ## png encoding
import math
from . import basic ## convert_svg_to_png
from .basic import MONOSPACE_FONT_FAMILY
from .html_colors import CB_RED, CB_GREEN, CB_BLUE, CB_ORANGE, CB_PURPLE


## these are not actually defined by the language...
all_text_height = {20:15.5, 100:75.}
all_text_width  = {20:12.5, 100:60.}

def rgb_from_fraction(fraction, cmap=None):

    assert fraction >=0 and fraction<=1

    fraction = float(fraction)

    if cmap is None:
        if fraction<0.5:
            i = 2*fraction*256

            #green = min(200,2*i)
            green = int( min(200,(200*i)/128) )
            blue = max(0,int( min(255,510-2*i) ))
            red = 0
        else:
            i = 2*(fraction-0.5)*256
            red = int( min(255,2*i) )
            green = int( max(0,min(200,400 - (200*i)/128)) )
            #green = min(200,510-2*i)
            blue = 0
    else:
        color_tuple = cmap(fraction)[:3]
        red, green, blue = [min(255,int(x*255)) for x in color_tuple]

    assert 0<=red  <=255
    assert 0<=green<=255
    assert 0<=blue <=255

    return '#{:02x}{:02x}{:02x}'.format(red,green,blue)


    # return '<rect x="{}" y="{}" height="{}" width="{}" fill="{}" stroke="{}"/>\n'\
    #     .format( upper_left[0], upper_left[1], lower_right[1]-upper_left[1], lower_right[0]-upper_left[0],
    #              fill, stroke )

def rectangle( upper_left, lower_right, fill, stroke, stroke_width=1, dashed=False ):
    stroke_dasharray_style= ';stroke-dasharray:5,5' if dashed else ''
    return '<rect x="{}" y="{}" height="{}" width="{}" style="fill:{};stroke:{};stroke-width:{}{}" />\n'\
        .format( upper_left[0], upper_left[1], lower_right[1]-upper_left[1], lower_right[0]-upper_left[0],
                 fill, stroke, stroke_width, stroke_dasharray_style )

def create_file( cmds, width, height, filename, create_png = False, background_color=None, use_xlink=False ):
    out = open( filename,'w')
    extra = '' if not use_xlink else 'xmlns:xlink="http://www.w3.org/1999/xlink"'
    out.write('<svg width="{}" height="{}" xmlns="http://www.w3.org/2000/svg" version="1.1" {} >\n'\
              .format(int(width),int(height),extra))
    if background_color:
        out.write(rectangle( (0,0), (width,height), background_color, 'white', 0 ) )
    out.write('\n'.join(cmds)+'\n')
    out.write('</svg>\n')
    out.close()

    if create_png:
        assert filename.endswith('.svg')
        pngfile = filename[:-4]+'.png'
        basic.convert_svg_to_png( filename, pngfile )


## this will work if oldfile was created using create_file
## returns: cmds, old_width, old_height
## cmds do not include the newline from the original file
## ie, they could go right into create_file
##
## matplotlib seems to generate files that are sized in pt units, which is a bit of a nuisance
## converting between them is something like #px = 1.25 * #pt
## pt = 1/72 inch, so inch = 72pt = 72*1.25=72*5/4 = 18*5 = 90px
## but somewhere on the web somebody says the new standard is 1inch=96px...
## at least that's what I get when I run convert on a matplotlib generated svg to make a png
## ( 576pt x 864pt goes to 720px x 1080px )
##
def embed_file( oldfile, x_shift, y_shift ):
    cmds = [] ##
    cmds = ['<g transform="translate({},{})" >'.format(x_shift,y_shift) ]
    for line in open(oldfile,'r'):
        if line.startswith("<svg"):
            l = line.split()
            assert l[1].startswith('width="')
            assert l[2].startswith('height="')
            old_width  = float( l[1].split('"')[1] ) ## will fail if pt or px or something in size
            old_height = float( l[2].split('"')[1] )
        elif line.startswith('</svg'):
            break
        else:
            cmds.append(line[:-1])
    cmds.append( '</g>')
    return cmds, old_width, old_height

## returns a single line
## note that we have to set use_xlink=True when we call create_file in order for the link to work here:
##
def embed_pngfile( pngfile, width, height, x0, y0, aspect=None ):
    assert exists( pngfile )
    encoded = base64.b64encode(open(pngfile, "rb").read()).decode('ascii')

    asp = '' if aspect==None else 'preserveAspectRatio="{}"'.format(aspect)

    cmd = '<image width="{}" height="{}" x="{}" y="{}" {} xlink:href="data:image/png;base64,{}" />'\
                     .format( width, height, x0, y0, asp, encoded )
    return cmd


def make_text( text, lower_left, fontsize,
               extra_tag = None, font_family = MONOSPACE_FONT_FAMILY, color = "black", font_weight = "normal" ):
    assert font_weight in ['normal','bold']
    cmd = '<text x="{:.3f}" y="{:.3f}" font-size="{}" font-weight="{}" font-family="{}" fill="{}" xml:space="preserve">{}</text>\n'\
        .format( lower_left[0], lower_left[1], fontsize, font_weight, font_family, color, text )
    return cmd


class SVG_tree_plotter:
    def __init__( self, cmap=None ):
        self.cmds = []
        self.cmap = cmap

    def make_line( self, p0, p1, line_width, normalized_score, extra_tag=None, color=None ):

        if normalized_score==None:
            #color='black'
            color = '#aaaaaa' # gray
        elif color==None:
            color = rgb_from_fraction(normalized_score, cmap=self.cmap)

        if p0[0] == p1[0]: ## vertical line, get width exactly right
            x = p0[0] ; y0 = min(p0[1],p1[1]) ; y1 = max(p0[1],p1[1] )

            upper_left = ( x-line_width/2.0, y0 )
            lower_right = ( x+line_width/2.0, y1 )
            cmd = rectangle( upper_left, lower_right, color, color )

        elif p0[1] == p1[1]: ## horizontal line, get width exactly right
            y = p0[1] ; x0 = min(p0[0],p1[0]) ; x1 = max(p0[0],p1[0] )

            upper_left  = ( x0, y-line_width/2.0 )
            lower_right = ( x1, y+line_width/2.0 )
            cmd = rectangle( upper_left, lower_right, color, color )
        else:
            cmd = '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" stroke-width="{}" stroke="{}"/>\n'\
                .format( p0[0], p0[1], p1[0], p1[1], 0.5*line_width,color )
        self.cmds.append( cmd )

    def make_text( self, text, p, fontsize, extra_tag=None ):
        color = 'black'
        ## shift so that p is the lower right
        textwidth = fontsize*0.6 * len(text)
        cmd = '<text x="{:.3f}" y="{:.3f}" font-size="{}" font-family="{}" fill="{}" xml:space="preserve">{}</text>\n'\
            .format( p[0]-textwidth, p[1], fontsize, MONOSPACE_FONT_FAMILY, color, text )
        self.cmds.append( cmd )

    def write( self, out ):
        out.write('\n'.join( self.cmds )+'\n' )



def color_stack( upper_left, lower_right, letters, colors, values ):
    fontsize= 20
    fontwidth  = all_text_width [fontsize]
    fontheight = all_text_height[fontsize]
    width  = lower_right[0] - upper_left[0]
    height = lower_right[1] - upper_left[1]

    ## draw them proportional to their values
    total = sum(values)

    ## draw from the top down
    height_sum = 0.
    lines = []

    for (letter,(color,value)) in zip( letters, list(zip( colors, values )) ):
        my_height = height * float(value)/total
        height_sum += my_height
        my_x = upper_left[0]
        my_y = upper_left[1] + height_sum

        ## we are going to have to scale
        x_scale = float( width )     / fontwidth
        y_scale = float( my_height ) / fontheight

        #print letter, 'my_height:',my_height, 'y_scale:',y_scale,'fontheight:',fontheight,'my_xy:',my_x,my_y

        ## we need to translate so that
        lines.append( '<text x="{:.6f}" y="{:.6f}" font-size="{}" font-family="{}" fill="{}" transform="scale({:.6f},{:.6f})">{}</text>\n'\
                      .format( my_x / x_scale, my_y / y_scale, fontsize, MONOSPACE_FONT_FAMILY, color,
                               x_scale, y_scale, letter ) )

        #lines.append( '<text x="{:.6f}" y="{:.6f}" font-size="{}" font-family="monospace" fill="{}" >{}</text>\n'\
        #              .format( my_x , my_y , fontsize, color, letter ) )
    return ''.join( lines )


def text_in_box( upper_left, lower_right, text, color ):
    fontsize= 100

    fontwidth  = all_text_width [fontsize]
    fontheight = all_text_height[fontsize]

    width  = lower_right[0] - upper_left[0]
    height = lower_right[1] - upper_left[1]

    ## we are going to have to scale
    x_scale = float( width )     / (fontwidth * len(text))
    y_scale = float( height ) / fontheight

    #print text, 'height:',height, 'y_scale:',y_scale,'fontheight:',fontheight,'my_xy:'

    ## we need to translate so that
    return '<text x="{:.6f}" y="{:.6f}" font-size="{}" font-family="{}" fill="{}" transform="scale({:.6f},{:.6f})">{}</text>\n'\
        .format( float(upper_left [0]) / x_scale,
                 float(lower_right[1]) / y_scale,
                 fontsize, MONOSPACE_FONT_FAMILY, color, x_scale, y_scale, text )



def protein_logo( upper_left, lower_right, pwm, scale={} ):## scale[pos] should be in range [0,1]

    aacolor = {}
    for aa in "G"       :aacolor[aa] = CB_ORANGE #"orange" ## special for glycine
    for aa in "STYC"    :aacolor[aa] = CB_GREEN #"green"
    for aa in "NQ"      :aacolor[aa] = CB_PURPLE #"purple"
    for aa in "KRH"     :aacolor[aa] = CB_BLUE #"blue"
    for aa in "DE"      :aacolor[aa] = CB_RED #"red"
    for aa in "P"       :aacolor[aa] = "black" ## for right now
    for aa in "AWFLIMV" :aacolor[aa] = "black" ## do we want to make W,F something different??
    for aa in "J":aacolor[aa] = "green"
    for aa in ".-":aacolor[aa] = "black"

    width  = lower_right[0] - upper_left[0]
    height = lower_right[1] - upper_left[1]

    N = len(list(pwm.keys()))

    letter_width = float(width) / N

    cmds = []

    for pos in range(N):
        l = [(y,x) for x,y in pwm[pos].items() ]
        l.sort()
        l.reverse()
        ## start from the top
        col_scale = scale.get(pos,1.0)
        assert col_scale>=0 and col_scale<=1.00001
        scaled_height = col_scale * height
        upper_y = lower_right[1] - scaled_height

        totfreq=0.0
        for freq,aa in l:
            if freq * scaled_height>1: ## at least one pixel high
                y0 = upper_y + totfreq * scaled_height
                y1 = upper_y + (totfreq+freq) * scaled_height
                x0 = upper_left[0] + pos*letter_width
                x1 = upper_left[0] + (pos+1)*letter_width
                cmds.append( text_in_box( (x0,y0), (x1,y1), aa, aacolor[aa] ) )
            totfreq += freq
            #print pos,aa,freq,totfreq,sum(pwm[pos].values()),pwm[pos]
        assert abs(totfreq-1.0)<1e-2

    return '\n'.join(cmds)+'\n'

def generic_logo( upper_left, lower_right, pwm ):

    color = 'black'

    width  = lower_right[0] - upper_left[0]
    height = lower_right[1] - upper_left[1]

    N = len(list(pwm.keys()))

    column_width = float(width) / N

    cmds = []

    for pos in range(N):
        l = [(y,x) for x,y in pwm[pos].items() ]
        l.sort()
        l.reverse()
        ## start from the top
        totfreq=0
        for freq,aa in l:
            if freq<1e-3:continue
            y0 = upper_left[1] + totfreq * height
            y1 = upper_left[1] + (totfreq+freq) * height
            x0 = upper_left[0] +  pos   *column_width
            x1 = upper_left[0] + (pos+1)*column_width
            cmds.append( text_in_box( (x0,y0), (x1,y1), aa, color ) )
            totfreq += freq
    return '\n'.join(cmds)+'\n'


def make_stack( upper_left, lower_right, l ):
    x0 = upper_left [0]
    x1 = lower_right[0]

    width  = lower_right[0] - upper_left[0]
    height = lower_right[1] - upper_left[1]

    l.sort()
    l.reverse()

    total = sum([x[0] for x in l])
    totval = 0
    cmds = []
    for vwc in l:
        if len(vwc) == 2:
            val,word = vwc
            color = 'black'
        else:
            assert len(vwc) == 3
            val,word,color = vwc
        y0 = upper_left[1] + height * float(totval) / total
        y1 = upper_left[1] + height * float(totval+val) / total
        cmds.append( text_in_box( (x0,y0), (x1,y1), word, color ) )
        totval+=val

    return '\n'.join(cmds)+'\n'



def enrichment_glyph_marker_old(marker_id):
    cmd = """<marker id="{}"
      viewBox="0 0 10 10" refX="10" refY="5"
      markerUnits="strokeWidth"
      markerWidth="4" markerHeight="3"
      orient="auto">
      <path d="M 0 0 L 10 5 L 0 10 z" />
    </marker>
    """.format(marker_id)
    return cmd

def enrichment_glyph_old( center, arrow_length, arrow_width, fontsize, marker_id, enrichment ):
    line_text_sep = 2.
    if enrichment>1: ## up arrow, text at the top
        line_p0 = [ center[0], center[1] + arrow_length/2. ]
        line_p1 = [ center[0], center[1] - arrow_length/2. ]
        text_lower_y = line_p1[1] - line_text_sep
        enrichment_text = '{:.1f}'.format(enrichment)
    else:
        line_p0 = [ center[0], center[1] - arrow_length/2. ]
        line_p1 = [ center[0], center[1] + arrow_length/2. ]
        text_lower_y = line_p1[1] + line_text_sep + 0.75 * fontsize
        enrichment_text = '{:.1f}'.format(1.0/enrichment)
    text_lower_x = center[0] - 0.6 * fontsize * 0.5 * len(enrichment_text)



    ## define the marker
    cmds = []
    cmds.append(
        '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" marker-end="url(#{})" stroke="black" stroke-width="{}"/>'\
        .format( line_p0[0], line_p0[1], line_p1[0], line_p1[1], marker_id, arrow_width ) )
    cmds.append( make_text( enrichment_text, [ text_lower_x, text_lower_y ], fontsize ) )

    return cmds

# eg_marker_id_prefix = 'eg_'
# eg_max_arrow_heads = 6

# def enrichment_glyph_markers():
#     global eg_marker_id_prefix
#     global eg_max_arrow_heads

#     for heads in range(1,eg_max_arrow_heads+1):
#         markerid = '{}{}'.format( eg_marker_id_prefix, heads )

#     cmd = """<marker id="{}"
#       viewBox="0 0 10 10" refX="10" refY="5"
#       markerUnits="strokeWidth"
#       markerWidth="4" markerHeight="3"
#       orient="auto">
#       <path d="M 0 0 L 10 5 L 0 10 z" />
#     </marker>
#     """.format(marker_id)
#     return cmd

def enrichment_glyph_cmds( center, arrow_length, arrow_width, enrichment, add_rectangle = False, box_color = 'gold' ):
    cmds = []

    num_heads = int(math.floor( abs( math.log(enrichment,2.) ) ))
    if num_heads<1:
        return cmds

    head_sep = 3. * arrow_width
    head_width = 3. * arrow_width
    head_slant = 1.

    if enrichment>1: ## up arrow, text at the top
        line_p0 = [ center[0], center[1] + arrow_length/2. ]
        line_p1 = [ center[0], center[1] - arrow_length/2. ]
        head_step = 1 * head_sep
    else:
        line_p0 = [ center[0], center[1] - arrow_length/2. ]
        line_p1 = [ center[0], center[1] + arrow_length/2. ]
        head_step = -1 * head_sep

    if add_rectangle:
        ## put a nice rounded rectangle around the outside
        rect_x0 = center[0] - head_width - arrow_width
        rect_y0 = center[1] - arrow_length/2 - arrow_width
        rect_width = 2*(head_width+arrow_width)
        rect_height = arrow_length + 2*arrow_width
        rect_round = 3.0 * arrow_width

        cmds.append( '<rect x="{:.3f}" y="{:.3f}" width="{:.3f}" height="{:.3f}" rx="{:.3f}" ry="{:.3f}" stroke="{}" fill="{}"/>'\
                     .format( rect_x0, rect_y0, rect_width, rect_height, rect_round, rect_round,
                              box_color, box_color ))


    ## make the line
    cmds.append( '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" stroke="black" stroke-width="{}"/>'\
                 .format( line_p0[0], line_p0[1], line_p1[0], line_p1[1], arrow_width ) )

    ## now make the heads

    for head in range(num_heads):
        for xsign in [1,-1]:
            x1 = line_p0[0]
            x2 = x1 + xsign * head_width
            y1 = line_p1[1] + head * head_step
            y2 = y1 + head_slant * head_step

            cmds.append( '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" stroke="black" stroke-width="{}"/>'\
                         .format( x1,y1,x2,y2, arrow_width ) )

            if False: ## add shifted white line
                y_shift = arrow_width * head_step / head_sep
                cmds.append( '<line x1="{:.3f}" y1="{:.3f}" x2="{:.3f}" y2="{:.3f}" stroke="white" stroke-width="{}"/>'\
                             .format( x1,y1+y_shift,x2,y2+y_shift, arrow_width ) )


    return cmds


if __name__ == '__main__':
    cmds = []
    # marker_id = 'tri'
    # cmds.append( enrichment_glyph_marker( marker_id ) )
    center = [50,50]
    arrow_length = 40.
    arrow_width = 3.
    enrichment = 2.01
    cmds += enrichment_glyph_cmds( center, arrow_length, arrow_width, enrichment )

    center = [100,50]
    enrichment = 45.01
    cmds += enrichment_glyph_cmds( center, arrow_length, arrow_width, enrichment )

    center = [250,50]
    enrichment = 0.49
    cmds += enrichment_glyph_cmds( center, arrow_length, arrow_width, enrichment )

    svgfile = 'tmp.svg'
    width = 300
    height = 100

    create_file( cmds, width, height, svgfile, create_png=True, background_color='red' )





    # out = open('tmp3.svg','w')

    # out.write( '<svg width="1000" height="1000">\n' )

    # texts = [ ( (0,0), (500,300), 'DUDE' ),\
    #               ( (100,350), (700,400), 'TEST' ) ]


    # for (upper_left,lower_right,text) in texts:


    #     out.write( '<rect x="{}" y="{}" height="{}" width="{}" fill="none" stroke="black"/>\n'\
    #                    .format( upper_left[0], upper_left[1], lower_right[1]-upper_left[1], lower_right[0]-upper_left[0] ) )

    #     out.write( text_in_box( upper_left, lower_right, text ) )


    # out.write( '</svg>\n' )
